from CommonTools.dtyper import _BaseDtype, _DtypeBase, \
    Array, Bool, DateT, DateTimeT, Decimal, Float, Integer, Map, String, \
    check_auto_cast_precision


class DtypeClickHouse(_DtypeBase):
    @classmethod
    def nullable_wrapper(cls, d: _BaseDtype, s: str) -> str:
        if d.nullable:
            s = f"Nullable({s})"
        return s

    @classmethod
    def toInteger(cls, d: Integer) -> str:
        precision = check_auto_cast_precision(
            d.precision,
            supported_precision=[8, 16, 32, 64, 128, 256]
        )
        base = f"Int{precision}"
        if not d.signed:
            base = "U" + base
        return cls.nullable_wrapper(d, base)

    @classmethod
    def toFloat(cls, d: Float) -> str:
        precision = check_auto_cast_precision(
            d.precision,
            supported_precision=[32, 64]
        )
        base = f"Float{precision}"
        return cls.nullable_wrapper(d, base)

    @classmethod
    def toDecimal(cls, d: Decimal) -> str:
        base = f"Decimal({d.precision},{d.scale})"
        return cls.nullable_wrapper(d, base)

    @classmethod
    def toBool(cls, d: Bool) -> str:
        return cls.nullable_wrapper(d, s = "Bool")

    @classmethod
    def toString(cls, d: String) -> str:
        return cls.nullable_wrapper(d, s = "String")

    @classmethod
    def toDateT(cls, d: DateT) -> str:
        return cls.nullable_wrapper(d, s = "Date32")

    @classmethod
    def toDateTimeT(cls, d: DateTimeT) -> str:
        if d.tz is not None:
            base = f"DateTime64({d.precision},'{d.tz}')"
        else:
            base = f"DateTime64({d.precision})"
        return cls.nullable_wrapper(d, s = base)

    @classmethod
    def toArray(cls, d: Array) -> str:
        e = d.element_dtype
        e.nullable = False  # TODO:  Array does not support nullable
        cls_name = e.__class__.__name__
        func = getattr(DtypeClickHouse, f"to{cls_name}")
        base = func(e)
        return f"Array({base})"

    @classmethod
    def toMap(cls, d: Map) -> str:
        k = d.key_dtype
        v = d.value_dtype
        k.nullable = False   # TODO:  key does not support nullable
        k_cls = k.__class__.__name__
        v_cls = v.__class__.__name__
        func_k = getattr(DtypeClickHouse, f"to{k_cls}")
        func_v = getattr(DtypeClickHouse, f"to{v_cls}")
        base_k = func_k(k)
        base_v = func_v(v)
        return f"Map({base_k},{base_v})"


from __future__ import annotations
import os
import shutil
from datetime import date
import time
from distutils.dir_util import copy_tree
from pathlib import Path
from typing import List
import logging
import pandas as pd
import pyspark
from tqdm.auto import tqdm

from CommonTools.SnapStructure.tools import get_file_list_pattern, match_date_from_str
from CommonTools.dbapi import baseDbInf, dbIO
from CommonTools.sparker import SparkConnector
from CommonTools.utils import str2dt

    
class SnapshotDataManagerBase:
    @classmethod
    def setup(cls, spark_connector: SparkConnector = None, default_engine:str = 'pandas'):
        if spark_connector:
            cls.sc = spark_connector
        cls.default_engine = default_engine

    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema , e.g., RAW, PROCESSED
        :param table:  the virtual table under each schema, e.g., CLNT_GENERAL, BDA_GENERAL
        """
        self.schema = schema
        self.table = table

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(schema={self.schema}; table={self.table})"
    
    def init_table(self, *args, **kws):
        """initialize schema/table

        :return:
        """
        raise NotImplementedError("")
    
    def get_schema(self) -> pd.Series:
        """Check dtypes of the given schema/dataset

        :return:
        """

        if self.default_engine == 'pandas':
            df_random = self.read_random_snap()
            return df_random.dtypes
        if self.default_engine == 'spark':
            df_random = self.load_random_snap()
            return pd.Series(dict(df_random.dtypes))
        else:
            raise ValueError("engine can only be spark or pandas")
    
    def pre_save_check(self, snap_dt: date, overwrite: bool = False) -> bool:
        """pre-check before inserting records, will delete existing records at snap_dt if overwrite is set to True

        :param snap_dt:
        :param overwrite: if False, will not insert record if exists; will delete otherwise
        :return: True = clear to go, False = exit in downstream
        """
        if overwrite:
            self.delete(snap_dt = snap_dt)
            return True
        else:
            num_records = self.count(snap_dt)
            if num_records > 0:
                logging.warning(f"Presave check failed, {num_records} existing records found for {self.schema}.{self.table}@{snap_dt} while overwrite is set to False")
                return False
            return True
    
    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        raise NotImplementedError("")
    
    def save(self, df: pd.DataFrame, snap_dt: date, overwrite: bool = False, **kws):
        """save pandas dataframe to the system

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :param overwrite snap_dt: whether to overwrite the existing record if any
        :raises NotImplementedError: _description_
        """
        if not self.pre_save_check(snap_dt = snap_dt, overwrite = overwrite):
            return
        
        self._save(df = df, snap_dt = snap_dt, **kws)
        logging.info(f"Successfully saved {len(df)} records to {self.schema}.{self.table}@{snap_dt}")

    def _save_partitions(self, df: pyspark.sql.DataFrame, snap_dt: date, **kws):
        """the pure logic to save pyspark dataframe to the system, without handling existing record problem

        :param pyspark.sql.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        raise NotImplementedError("")

    def save_partitions(self, df: pyspark.sql.DataFrame, snap_dt: date, overwrite: bool = False, **kws):
        """save pyspark dataframe to the system

        :param pyspark.sql.DataFrame df: _description_
        :param date snap_dt: _description_
        :param overwrite snap_dt: whether to overwrite the existing record if any
        :raises NotImplementedError: _description_
        """
        if not self.pre_save_check(snap_dt = snap_dt, overwrite = overwrite):
            return
        
        self._save_partitions(df = df, snap_dt = snap_dt, **kws)
        logging.info(f"Successfully saved {df.count()} records to {self.schema}.{self.table}@{snap_dt}")

    def etl_from_db(self, db_conf: baseDbInf, sql: str, snap_dt: date, engine: str = 'pandas'):
        """do ETL work, extracting data from database and save it to local dir

        :param db_conf: the database configuration object, specify which db and connection param
        :param sql: the sql to do the ETL
        :param snap_dt: the snap date to do the ETL, for storage purpose
        :param engine: the engine to query_extract {pandas, spark}
        :return:
        """
        if engine == "pandas":
            db_conn = dbIO(db_conf)
            df = db_conn.query_sql_df(sql)
            self.save(df, snap_dt)
        elif engine == "spark":
            # spark data query_extract
            if hasattr(self, "sc"):
                df = self.sc.query_db(db_conf, sql)
                self.save_partitions(df, snap_dt)
            else:
                raise ValueError("No spark session binded, please bind via .setup() method")
        else:
            raise ValueError("Engine can only be pandas or spark")
        
    def get_existing_snap_dts(self) -> List[date]:
        raise NotImplementedError("")
    
    def count(self, snap_dt: date) -> int:
        """count the sample size for given snap date

        :param date snap_dt: _description_
        :return int: # of rows of the dataframe
        """
        if snap_dt not in self.get_existing_snap_dts():
            return 0
        if self.default_engine == 'pandas':
            df = self.read(snap_dt=snap_dt)
            return len(df)
        if self.default_engine == 'spark':
            df = self.load(snap_dt=snap_dt)
            return df.count()
        else:
            raise ValueError("engine can only be spark or pandas")
    
    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        raise NotImplementedError("")
    
    def reads(self, snap_dts: List[date], **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        raise NotImplementedError("")

    def read_random_snap(self) -> pd.DataFrame:
        """Select a random snap from the schema.table

        :return:
        """
        import random

        existing_snaps = self.get_existing_snap_dts()
        random_snap = random.choice(existing_snaps)
        return self.read(random_snap)
    
    def load(self, snap_dt: date, **kws) -> pyspark.sql.DataFrame:
        """Read as spark dataframe (one snapshot date) data, and can also access from sc temporary view

        :param snap_dt: snap_dt to load
        :return:
        """
        raise NotImplementedError("")

    def load_random_snap(self) -> pyspark.sql.DataFrame:
        """Select a random snap from the schema.table

        :return:
        """
        import random

        existing_snaps = self.get_existing_snap_dts()
        random_snap = random.choice(existing_snaps)
        return self.load(random_snap)
    
    def loads(self, snap_dts: List[date], **kws) -> pyspark.sql.DataFrame:
        """reads as pyspark dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        raise NotImplementedError("")
    
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        raise NotImplementedError("")

    def drop(self):
        """drop the whole table

        :return:
        """
        raise NotImplementedError("")

    def migrate(self, dst_schema: str, dst_table: str):
        """move from the existing schema.table to new one, the existing one will be deleted

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        new = self.duplicate(dst_schema, dst_table)
        self.drop()
        return new

    def duplicate(self, dst_schema: str, dst_table: str):
        """duplicate the existing schema.table to new one, the existing one will be kept

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        raise NotImplementedError("")

    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date on disk

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        raise NotImplementedError("")
    
    def is_valid(self, snap_dt, rows_threshold: int = 100) -> bool:
        rows = self.count(snap_dt = snap_dt)
        return rows > rows_threshold
    


class SnapshotDataManagerLocalFS(SnapshotDataManagerBase):
    """files are saved as parquet snapshots under each schema.table
        file naming convention: dir/tablename_YYYY-MM-DD.parquet
    """

    @classmethod
    def setup(cls, root_dir: str, spark_connector: SparkConnector = None, default_engine:str = 'pandas'):
        super(SnapshotDataManagerLocalFS, cls).setup(
            spark_connector = spark_connector,
            default_engine = default_engine
        )
        cls.ROOT_DIR = root_dir
        cls.default_engine = default_engine
        
    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema , e.g., RAW, PROCESSED
        :param table:  the virtual table under each schema, e.g., CLNT_GENERAL, BDA_GENERAL
        """
        super().__init__(schema = schema, table = table)
        dir = os.path.join(self.ROOT_DIR, schema, table)
        self.dir = dir  # the root path of the table under each schema
        self.init_table()

    def init_table(self):
        # https://note.nkmk.me/en/python-os-mkdir-makedirs/
        os.makedirs(self.dir, exist_ok = True)

    def get_filename(self, snap_dt: date) -> str:
        return f"{self.table}_{snap_dt}.parquet"

    def get_default_file_path(self, snap_dt: date) -> str:
        return os.path.join(self.dir, self.get_filename(snap_dt))
    

    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        filepath = self.get_default_file_path(snap_dt)
        df.to_parquet(filepath, **kws)
        logging.info(f"Successfully saved file to {filepath}")

    def _save_partitions(self, df: pyspark.sql.DataFrame, snap_dt: date, **kws):
        """the pure logic to save pyspark dataframe to the system, without handling existing record problem

        :param pyspark.sql.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        filepath = self.get_default_file_path(snap_dt)
        self.sc.save_parquet(df, filepath, **kws)
        logging.info(f"Successfully saved file to {filepath}")

    def get_existing_snap_dts(self) -> List[date]:
        existing_files = get_file_list_pattern(
            self.dir,
            pattern = f"*{self.table}*",
            include_dir = False  # don't include directory path
        )

        existing_snaps = list(set(str2dt(match_date_from_str(f)) for f in existing_files))
        existing_snaps.sort()
        return existing_snaps

    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_default_file_path(snap_dt)
        return pd.read_parquet(filepath, **kws)

    def reads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        # do the checks to ensure successful concatenation
        li =  []
        columns = pd.Index([])
        for snap_dt in tqdm(snap_dts):
            df_partition = self.read(snap_dt=snap_dt, **kws)
            if partition_key is not None and partition_key not in df_partition.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found in {snap_dt}")
            if len(columns) > 0 and len(columns.union(df_partition.columns)) != len(columns):
                # if new column set has a different length than the previous one, means the columns are not same across all the table partitions
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent for snapshot {snap_dt}")
            li.append(df_partition)

        return pd.concat(li, axis = 0, ignore_index = True)


    def load(self, snap_dt: date, **kws) -> pyspark.sql.DataFrame:
        """Read as spark dataframe (one snapshot date) data, and can also access from sc temporary view

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_default_file_path(snap_dt)
        if hasattr(self, "sc"):
            df = self.sc.read_parquet(filepath, **kws)
            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")

    def loads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pyspark.sql.DataFrame:
        """reads as pyspark dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        file_paths = [self.get_default_file_path(snap_dt) for snap_dt in snap_dts]
        if hasattr(self, "sc"):
            df = self.sc.read_parquet(*file_paths, **kws) # spark will automatically stack all tables together
            # do the partition check
            if partition_key is not None and partition_key not in df.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found")

            # pick up random snapshot data to check the columns are consistent
            one_columns = self.load_random_snap().columns
            if len(one_columns) != len(pd.Index(one_columns).union(df.columns)):
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent in one of the snapshot data")

            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")

    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        filepath = self.get_default_file_path(snap_dt)
        if os.path.exists(filepath):
            if os.path.isdir(filepath):
                # parquet partitions
                shutil.rmtree(filepath, ignore_errors = True)
            else:
                os.remove(filepath)

    def drop(self):
        """drop the whole table

        :return:
        """
        shutil.rmtree(self.dir)

    def migrate(self, dst_schema: str, dst_table: str):
        """move from the existing schema.table to new one, the existing one will be deleted

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        new = self.duplicate(dst_schema, dst_table)
        self.drop()
        return new

    def duplicate(self, dst_schema: str, dst_table: str):
        new = SnapshotDataManagerLocalFS(schema = dst_schema, table = dst_table)
        copy_tree(self.dir, new.dir)
        return new

    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date file (pandas) or folder (pyspark partitions)

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        f = self.get_default_file_path(snap_dt)
        if os.path.isfile(f):
            size_bytes = os.path.getsize(f)
        else:
            size_bytes = sum(p.stat().st_size for p in Path(f).rglob('*'))

        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size

    def is_valid(self, snap_dt, size_threshold: float = 0.5) -> bool:
        """Check if the file is valid
        
        """
        f = self.get_default_file_path(snap_dt)
        if os.path.isdir(f):
            # if it is a folder (partition files), it should has a .__SUCCESS.crc file generated by pyspark
            return "._SUCCESS.crc" in os.listdir(f)
        else:
            # if a single file, the size must be exceed the valid threshold
            return self.disk_space(snap_dt, unit = 'MB') > size_threshold


from configparser import ConfigParser
from getpass import getpass
from typing import List

from CommonTools.settings import SETTINGS

def get_credential(section: str, attr_keys: List[str] = None):
    """get credentials from credential file specified in Settings

    :param section: which section in the .ini file will extract the credentials
    :param attr_keys: a list of keys, by default will be [username, password]
    :return:
    """
    if not attr_keys:
        attr_keys = ['username', 'password']

    if SETTINGS.DB_CREDENTIAL_INTERACTIVE:
        values = []
        for attr in attr_keys:
            v = getpass(prompt = attr)
            values.append(v)
        return values
    else:
        config = ConfigParser()
        config.read(SETTINGS.DB_CREDENTIAL_PATH)
        sec = config[section]
        values = []
        for attr in attr_keys:
            v = sec[attr]
            values.append(v)
        return values

if __name__ == '__main__':
    SETTINGS.SET_CREDENTIAL_INTERACTIVE()



# different from data dtype from data engineering perspective, each varaible should also have analytical dtype
from collections import OrderedDict
from typing import Dict, List

from CommonTools.dtyper import _Dtype


class _BaseAnlDtype:
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        self.dtype = dtype
        self.dimensional = dimensional # if it is static attribute (should not change over time)

class IdPolicyAnlDtype(_BaseAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype=dtype, dimensional=dimensional)

class BinaryAnlDtype(_BaseAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype=dtype, dimensional=dimensional)

class NumericAnlDtype(_BaseAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype = dtype, dimensional = dimensional)

class CardinalNumAnlDtype(NumericAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype = dtype, dimensional = dimensional)

class OrdinalNumAnlDtype(NumericAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype = dtype, dimensional = dimensional)


class CategoricalAnlDtype(_BaseAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype=dtype, dimensional=dimensional)

class OrderedCategoricalAnlDtype(CategoricalAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype=dtype, dimensional=dimensional)

class UnOrderedCategoricalAnlDtype(CategoricalAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype=dtype, dimensional=dimensional)

class DateAnlDtype(_BaseAnlDtype):
    def __init__(self, dtype: _Dtype = None, dimensional: bool = False):
        super().__init__(dtype=dtype, dimensional=dimensional)


class AnalyticalSchema(OrderedDict):
    def __init__(self, schema_mapper: Dict[str, _BaseAnlDtype]):
        """create Schema object from predefined schema

        :param schema_mapper: {feature_name, feature_type} type should use dtyper._Dtype
        """
        super().__init__(**schema_mapper)

    def getColsByAnlDtype(self, anlDtype: type) -> List[str]:
        return list(filter(lambda t: isinstance(self[t], anlDtype), self.keys()))





from datetime import date
import logging
from typing import List
import pandas as pd
import pyspark
from CommonTools.sparker import SparkConnector
from CommonTools.utils import dt2str, str2dt
from CommonTools.SnapStructure.structure import SnapshotDataManagerBase
from ProviderTools.clickhouse.dbapi import ClickHouse, ClickhouseCRUD


class SnapshotDataManagerCHSQL(SnapshotDataManagerBase):
    """files are saved as clickhouse tables under each schema.table
    """

    @classmethod
    def setup(cls, ch_conf: ClickHouse, spark_connector: SparkConnector = None):
        super(SnapshotDataManagerCHSQL, cls).setup(spark_connector = spark_connector)
        cls.ch_conf = ch_conf
        cls.ch = ClickhouseCRUD(ch_conf)

    def __init__(self, schema:str, table:str, snap_dt_key: str):
        """database management interface

        :param schema: schema , e.g., RAW, PROCESSED
        :param table:  table under each schema, e.g., CLNT_GENERAL, BDA_GENERAL
        :param snap_dt_key: snap date column name for all tables
        """
        super().__init__(schema = schema, table = table)
        self.snap_dt_key = snap_dt_key

    def is_exist(self) -> bool:
        return self.ch.is_exist(schema=self.schema, table=self.table)

    def init_table(self, primary_keys:List[str], col_schemas: pd.DataFrame, table_note:str = None, overwrite:bool = False):
        if self.is_exist():
            if overwrite:
                self.ch.drop_table(schema=self.schema, table=self.table)
            else:
                logging.warning(f"{self.schema}.{self.table} already exists, will do nothing. set overwrite to True if you wish to reset table")
                return
        
        self.ch.create_schema(
            schema = self.schema
        )
        self.ch.create_table(
            schema = self.schema,
            table = self.table,
            col_schemas = col_schemas,
            engine = "MergeTree",
            partition_keys = [self.snap_dt_key],
            primary_keys = primary_keys,
            order_keys = primary_keys,
            table_note = table_note
        )

    def get_schema(self) -> pd.Series:
        """Check dtypes of the given schema/dataset

        :return:
        """
        schema = self.ch.detect_table_schema(self.schema, self.table)
        return pd.Series(schema['type'].tolist(), index = schema['name'].tolist(), name = 'dtype')

    def count(self, snap_dt: date) -> int:
        sql = "SELECT COUNT() FROM {schema:Identifier}.{table:Identifier} WHERE {partition_key:Identifier} = {snap_dt:Date32}"
        args = dict(schema=self.schema, table=self.table, partition_key = self.snap_dt_key, snap_dt=snap_dt)
        return self.ch.client.command(cmd = sql, parameters = args)

    def save_qry(self, query: str, snap_dt: date, overwrite: bool = False, keep_view:bool = True, **kws):
        """save the query into the table using "insert into schema.table select ... " clause

        :param query: the sql query for table to be inserted
        :param snap_dt: the snap date partition to be inserted, use to check existence
        :param overwrite: whether to overwrite the table if exists
        :param keep_view: the process will create temp view first, whether to keep that temp view
        :return:
        """
        # first detect whether snap date already exists
        clear = self.pre_save_check(snap_dt = snap_dt, overwrite = overwrite)
        if clear is False:
            return

        # first create a view of that table which can be inspected the column dtypes and names
        dt_str = dt2str(snap_dt, format = "%Y%m")
        view_name = f"{self.table}_V_{dt_str}"
        self.ch.create_view(schema = self.schema, view = view_name, query = query)
        qry_cols = self.ch.detect_table_schema(schema = self.schema, table = view_name)['name'].tolist()

        # insert into the table
        self.ch.insert_from_select(
            schema = self.schema,
            table = self.table,
            query = query,
            qry_cols = qry_cols # infered from view
        )

        # delete the view if required
        if not keep_view:
            self.ch.drop_view(schema = self.schema, view = view_name)

        logging.info(f"Successfully saved to {self.schema}.{self.table}@{snap_dt}")

    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        self.ch.client.insert_df(
            table = self.table,
            database = self.schema,
            df = df,
            **kws
        )

    def ingest_from_file(self, file_path:str, snap_dt: date, header:bool = True, overwrite:bool = True):
        if overwrite:
            self.delete(snap_dt = snap_dt)
        else:
            num_records = self.count(snap_dt)
            if num_records > 0:
                logging.warning(f"{num_records} records found for {self.schema}.{self.table}@{snap_dt}, will do nothing")
                return

        self.ch.ingest_file(
            schema = self.schema,
            table = self.table,
            file_path = file_path,
            header = header,
            use_cli = False
        )

    def get_existing_snap_dts(self) -> List[date]:
        sql = f"""
        select distinct {self.snap_dt_key}
        from {self.schema}.{self.table}
        order by {self.snap_dt_key}
        """
        existing_snaps = self.ch.client.query_df(sql)
        if len(existing_snaps) == 0:
            return []
        return list(str2dt(dt.date()) for dt in existing_snaps[self.snap_dt_key].dt.to_pydatetime())

    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        if 'columns' in kws:
            cols = ','.join(kws['columns'])
        else:
            cols = '*'
        sql = f"""
        select {cols} from {self.schema}.{self.table} where {self.snap_dt_key} = '{snap_dt}'
        """
        return self.ch.client.query_df(sql)

    def reads(self, snap_dts: List[date], **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        if 'columns' in kws:
            cols = ','.join(kws['columns'])
        else:
            cols = '*'
        snap_dt_range = ",".join(f"'{dt}'" for dt in snap_dts)
        sql = f"""
        select {cols} from {self.schema}.{self.table} where {self.snap_dt_key} in [{snap_dt_range}]
        """
        return self.ch.client.query_df(sql)


    def load(self, snap_dt: date, **kws) -> pyspark.sql.DataFrame:
        """Read as spark dataframe (one snapshot date) data, and can also access from sc temporary view

        :param snap_dt: snap_dt to load
        :return:
        """
        if 'columns' in kws:
            cols = ','.join(kws['columns'])
        else:
            cols = '*'
        sql = f"""
        select {cols} from {self.schema}.{self.table} where {self.snap_dt_key} = '{snap_dt}'
        """
        if hasattr(self, "sc"):
            df = self.sc.query_db(self.ch_conf, sql)
            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")

    def loads(self, snap_dts: List[date], **kws) -> pyspark.sql.DataFrame:
        """reads as pyspark dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        if 'columns' in kws:
            cols = ','.join(kws['columns'])
        else:
            cols = '*'
        snap_dt_range = ",".join(f"'{dt}'" for dt in snap_dts)
        sql = f"""
        select {cols} from {self.schema}.{self.table} where {self.snap_dt_key} in [{snap_dt_range}]
        """
        if hasattr(self, "sc"):
            df = self.sc.query_db(self.ch_conf, sql)
            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")


    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        self.ch.delete_data(schema=self.schema, table=self.table, where_clause = f"{self.snap_dt_key} = '{snap_dt}'")

    def drop(self):
        """drop the whole table

        :return:
        """
        self.ch.drop_table(schema=self.schema, table=self.table)

    def duplicate(self, dst_schema: str, dst_table: str):
        sql = """insert into {dst_schema:Identifier}.{dst_table:Identifier} select * from {src_schema:Identifier}.{src_table:Identifier}"""
        args = dict(dst_schema = dst_schema, dst_table = dst_table, src_schema = self.schema, src_table = self.table)
        return self.ch.client.command(cmd = sql, parameters = args)

    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date file (pandas) or folder (pyspark partitions)

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        sql = """
        select 
            sum(rows) as rows,
            sum(bytes_on_disk) as bytes_on_disk
        from system.parts
        where 
            active
            and database = %(schema)s
            and table = %(table)s
            and partition = %(snap_dt)s
        """
        args = dict(schema = self.schema, table =self.table, snap_dt = snap_dt)
        d = self.ch.client.query(query=sql, parameters=args).first_item

        size_bytes, rows = d['bytes_on_disk'], d['rows']
        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size

    def is_valid(self, snap_dt, rows_threshold: int = 0) -> bool:
        """Check if the file is valid"""
        sql = f"""
        select count() as size from {self.schema}.{self.table} where {self.snap_dt_key} = '{snap_dt}'
        """
        rows = self.ch.client.query(query = sql).first_item['size']
        return rows > rows_threshold


import logging
from collections import OrderedDict
from enum import Enum
from typing import Dict
import yaml
import numpy as np


#### Internal representation types
class _Dtype:
    def collect_args(self) -> dict:
        return {}

    def to_config(self) -> dict:
        return {
            "type": self.__class__.__name__,
            "args": self.collect_args()
        }

    @classmethod
    def from_config(cls, config: dict):
        template = config['type']
        params = config['args']
        mapping = {
            'bool' : Bool,
            'integer' : Integer,
            'float' : Float,
            'decimal' : Decimal,
            'string' : String,
            'datet' : DateT,
            'datetimet': DateTimeT,
            'array': Array,
            'map': Map
        }
        block = mapping.get(template.lower()) # the class of relavant dtype
        return block.from_args(params)

class _BaseDtype(_Dtype):
    def __init__(self, nullable:bool = True):
        self.nullable = bool(nullable)

    @classmethod
    def from_args(cls, args:dict):
        # de-serialize
        return cls(**args)

    def __repr__(self):
        return self.__str__()

    def collect_args(self) -> dict:
        return dict(nullable = self.nullable)

class Integer(_BaseDtype):
    def __init__(self, nullable:bool = True, precision: int = 32, signed:bool = True):
        super().__init__(nullable)
        self.precision = int(precision)
        self.signed = bool(signed)

    def __str__(self):
        # will use numpy type
        return f"Integer{self.precision}(null={self.nullable}, signed={self.signed})"

    def collect_args(self) -> dict:
        return dict(nullable = self.nullable, precision = self.precision, signed = self.signed)


class Decimal(_BaseDtype):
    def __init__(self, nullable: bool = True, precision: int = 10, scale: int = 2):
        super().__init__(nullable)
        self.precision = int(precision)
        self.scale = int(scale)

    def __str__(self):
        return f"Decimal({self.precision}, {self.scale}, null={self.nullable})"

    def collect_args(self) -> dict:
        return dict(nullable = self.nullable, precision = self.precision, scale = self.scale)

class Float(_BaseDtype):
    def __init__(self, nullable: bool = True, precision: int = 32):
        super().__init__(nullable)
        self.precision = int(precision)

    def __str__(self):
        return f"Float{self.precision}(null={self.nullable})"

    def collect_args(self) -> dict:
        return dict(nullable = self.nullable, precision = self.precision)

class Bool(_BaseDtype):
    def __str__(self):
        return f"Bool(null={self.nullable})"

class String(_BaseDtype):
    def __str__(self):
        return f"String(null={self.nullable})"

class DateT(_BaseDtype):
    def __str__(self):
        return f"DateT(null={self.nullable})"

class DateTimeT(_BaseDtype):
    def __init__(self, nullable: bool = True, precision:float = 0, tz: str = None):
        super().__init__(nullable)
        self.precision = precision
        self.tz = str(tz)  # whether timezone aware

    def __str__(self):
        return f"DateTimeT(null={self.nullable}, precision = {self.precision}, tz={self.tz})"

    def collect_args(self) -> dict:
        return dict(nullable = self.nullable, precision = self.precision, tz = self.tz)

class _ComplexDtype(_Dtype):
    pass

class Array(_ComplexDtype):
    def __init__(self, element_dtype: _Dtype):
        self.element_dtype = element_dtype

    def __str__(self):
        return f"Array<{self.element_dtype}>"

    @classmethod
    def from_args(cls, args: dict):
        # de-serialize
        element_config = args['element_dtype']
        return cls(element_dtype = _Dtype.from_config(element_config))

    def collect_args(self) -> dict:
        return dict(element_dtype = self.element_dtype.to_config())

class Map(_ComplexDtype):
    def __init__(self, key_dtype: _BaseDtype, value_dtype: _Dtype):
        self.key_dtype = key_dtype
        self.value_dtype = value_dtype

    def __str__(self):
        return f"Map<{self.key_dtype},{self.value_dtype}>"

    @classmethod
    def from_args(cls, args: dict):
        # de-serialize
        key_config = args['key_dtype']
        value_config = args['value_dtype']

        return cls(
            key_dtype = _Dtype.from_config(key_config),
            value_dtype = _Dtype.from_config(value_config)
        )

    def collect_args(self) -> dict:
        return dict(key_dtype = self.key_dtype.to_config(), value_dtype = self.value_dtype.to_config())

##### language specific interface

def check_auto_cast_precision(precision: int, supported_precision: list, force_cap: bool = False):
    if precision in supported_precision:
        # pass check
        return precision
    # if not, will cast to a bigger precision
    idx = np.searchsorted(supported_precision, precision, side='left')
    if idx >= len(supported_precision):
        if force_cap:
            logging.warning(f"Max precision exceeded, will force precision cap to {max(supported_precision)}")
            return max(supported_precision)
        raise TypeError(
            f"Precision Overflow, max support precision: {max(supported_precision)}, provided precision: {precision}")
    return supported_precision[idx]

class _DtypeBase:
    @classmethod
    def translate(cls, d: _Dtype) -> str:
        # auto detect what type is
        cls_name = d.__class__.__name__
        func = getattr(cls, f"to{cls_name}")
        return func(d)

    @classmethod
    def toInteger(cls, d: Integer) -> str:
        raise NotImplementedError

    @classmethod
    def toDecimal(cls, d: Decimal) -> str:
        raise NotImplementedError

    @classmethod
    def toFloat(cls, d: Float) -> str:
        raise NotImplementedError

    @classmethod
    def toBool(cls, d: Bool) -> str:
        raise NotImplementedError

    @classmethod
    def toString(cls, d: String) -> str:
        raise NotImplementedError

    @classmethod
    def toDateT(cls, d: DateT) -> str:
        raise NotImplementedError

    @classmethod
    def toDateTimeT(cls, d: DateTimeT) -> str:
        raise NotImplementedError

    @classmethod
    def toArray(cls, d: Array) -> str:
        raise NotImplementedError

    @classmethod
    def toMap(cls, d: Map) -> str:
        raise NotImplementedError

class DtypePandas(_DtypeBase):
    @classmethod
    def toInteger(cls, d: Integer) -> str:
        precision = check_auto_cast_precision(
            d.precision,
            supported_precision=[8, 16, 32, 64]
        )
        if d.nullable:
            # will use pandas extended type
            base = f"Int{precision}"
            if not d.signed:
                base = "U" + base
        else:
            # will use numpy type
            base = f"int{precision}"
            if not d.signed:
                base = "u" + base

        return base

    @classmethod
    def toDecimal(cls, d: Decimal) -> str:
        return "float64"

    @classmethod
    def toFloat(cls, d: Float) -> str:
        precision = check_auto_cast_precision(
            d.precision,
            supported_precision=[32, 64]
        )
        return f"float{precision}"

    @classmethod
    def toBool(cls, d: Bool) -> str:
        if d.nullable:
            return "boolean" # pandas extended type - nullable
        return "bool" # pandas original type

    @classmethod
    def toString(cls, d: String) -> str:
        return "string"

    @classmethod
    def toDateT(cls, d: DateT) -> str:
        # no date/datetime diff
        return "datetime64[ns]"

    @classmethod
    def toDateTimeT(cls, d: DateTimeT) -> str:
        if d.tz is not None:
            return f"datetime64[ns, {d.tz}]" # timezone aware type
        return "datetime64[ns]"

class DtypeSparkSQL(_DtypeBase):
    @classmethod
    def toInteger(cls, d: Integer) -> str:
        precision = check_auto_cast_precision(
            d.precision,
            supported_precision=[8, 16, 32, 64]
        )
        return {
            8: 'Byte',
            16: 'Short',
            32: 'Integer',
            64: 'Long'
        }.get(precision)

    @classmethod
    def toDecimal(cls, d: Decimal) -> str:
        return f"Decimal({d.precision},{d.scale})"

    @classmethod
    def toFloat(cls, d: Float) -> str:
        precision = check_auto_cast_precision(
            d.precision,
            supported_precision=[32, 64]
        )
        return {
            32: 'Float',
            64: 'Double'
        }.get(precision)

    @classmethod
    def toBool(cls, d: Bool) -> str:
        return "Boolean"

    @classmethod
    def toString(cls, d: String) -> str:
        return "String"

    @classmethod
    def toDateT(cls, d: DateT) -> str:
        return "Date"

    @classmethod
    def toDateTimeT(cls, d: DateTimeT) -> str:
        return "TimeStamp"

    @classmethod
    def toArray(cls, d: Array) -> str:
        e = d.element_dtype
        cls_name = e.__class__.__name__
        func = getattr(DtypeSparkSQL, f"to{cls_name}")
        base = func(e)
        return f"Array<{base}>"

    @classmethod
    def toMap(cls, d: Map) -> str:
        k = d.key_dtype
        v = d.value_dtype
        k_cls = k.__class__.__name__
        v_cls = v.__class__.__name__
        func_k = getattr(DtypeSparkSQL, f"to{k_cls}")
        func_v = getattr(DtypeSparkSQL, f"to{v_cls}")
        base_k = func_k(k)
        base_v = func_v(v)
        return f"Map<{base_k},{base_v}>"


######### Schema handling -- batch conversion

class Schema(OrderedDict):
    def __init__(self, schema_mapper: Dict[str, _Dtype]):
        """create Schema object from predefined schema

        :param schema_mapper: {feature_name, feature_type} type should use dtyper._Dtype
        """
        super().__init__(**schema_mapper)

    def to_js(self) -> OrderedDict:
        r = OrderedDict()
        for col, dtype in self.items():
            r[col] = dtype.to_config()
        return r
    
    def to_yaml(self, filename: str):
        with open(filename, 'w') as obj:
            yaml.dump(
                self.to_js(), 
                obj,
                default_flow_style=False, 
                sort_keys=False
            )

    @classmethod
    def from_js(cls, js: OrderedDict):
        """create Schema object from dictionary (json format)

        :param schema:
        :return: Schema()
        """
        s = Schema({})
        for col, config in js.items():
            s[col] = _Dtype.from_config(config)
        return s
    
    @classmethod
    def from_yaml(cls, filename: str):
        with open(filename, 'r') as obj:
            tt = yaml.load(obj, Loader=yaml.Loader)
            schemas = cls.from_js(tt)
        return schemas



if __name__ == '__main__':
    i = Integer(nullable=True, precision=64, signed=False)
    print(i)
    print(DtypeSparkSQL.toInteger(i), DtypeSparkSQL.translate(i))

    f = Array(Integer(nullable=True, precision=64, signed=False))
    print(f)
    print(DtypeSparkSQL.toArray(f), DtypeSparkSQL.translate(f))

    m = Map(
        String(nullable=False),
        Integer(nullable=True, precision=64, signed=False),
    )
    print(m)

    print(i.to_config())
    print(m.to_config())
    print(Array.from_config(f.to_config()))
    print(Map.from_config(m.to_config()))


import os
from datetime import date
import logging
from typing import List
import pandas as pd
import pyspark
from tqdm.auto import tqdm

from CommonTools.sparker import SparkConnector
from CommonTools.utils import dt2str, str2dt
from CommonTools.SnapStructure.structure import SnapshotDataManagerBase
from CommonTools.SnapStructure.tools import match_date_from_str
from ProviderTools.aws.s3 import S3Accessor


class SnapshotDataManagerS3(SnapshotDataManagerBase):
    """files are saved as s3 objs under each schema.table
    """

    @classmethod
    def setup(cls, bucket: str,  root_dir: str, s3a: S3Accessor, spark_connector: SparkConnector = None, default_engine:str = 'pandas'):
        super(SnapshotDataManagerS3, cls).setup(
            spark_connector = spark_connector,
            default_engine = default_engine
        )
        cls.bucket = bucket
        cls.ROOT_DIR = root_dir if not root_dir.startswith("/") else root_dir[1:]
        cls.s3a = s3a
        cls.s3a.enter_bucket(bucket_name=bucket)
    
    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema , e.g., RAW, PROCESSED
        :param table:  the virtual table under each schema, e.g., CLNT_GENERAL, BDA_GENERAL
        """
        super().__init__(schema = schema, table = table)
        dir = os.path.join(self.ROOT_DIR, schema, table)
        self.dir = dir  # the root path of the table under each schema

    def get_filename(self, snap_dt: date) -> str:
        return f"{self.table}_{snap_dt}.parquet"

    def get_default_file_path(self, snap_dt: date) -> str:
        return os.path.join(self.dir, self.get_filename(snap_dt))
    
    def get_file_path_with_s3_prefix(self, snap_dt: date) -> str:
        return f"s3a://{self.bucket}/{self.get_default_file_path(snap_dt=snap_dt)}"

    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        self.s3a.save_parquet(
            df = df,
            remote_path = self.get_default_file_path(snap_dt=snap_dt),
            **kws
        )
    
    def _save_partitions(self, df: pyspark.sql.DataFrame, snap_dt: date, **kws):
        """the pure logic to save pyspark dataframe to the system, without handling existing record problem

        :param pyspark.sql.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        raise NotImplementedError("")
    
    def get_existing_snap_dts(self) -> List[date]:
        df = self.s3a.list_objs(remote_path = self.dir)
        existing_snaps = list(set(map(
            str2dt,
            (
                df
                .loc[
                    (df['Key'].str.startswith(pat = f"{self.dir}/"))
                    & (df['Key'].str.endswith(pat = ".parquet")), 
                    'Key']
                .str.replace(self.dir, "")
                .apply(match_date_from_str)
                .tolist()
            )
        )))
        existing_snaps.sort()
        return existing_snaps
    
    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_default_file_path(snap_dt=snap_dt)
        return self.s3a.read_parquet(remote_path = filepath, **kws)
    

    def reads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        # do the checks to ensure successful concatenation
        li =  []
        columns = pd.Index([])
        for snap_dt in tqdm(snap_dts):
            df_partition = self.read(snap_dt=snap_dt, **kws)
            if partition_key is not None and partition_key not in df_partition.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found in {snap_dt}")
            if len(columns) > 0 and len(columns.union(df_partition.columns)) != len(columns):
                # if new column set has a different length than the previous one, means the columns are not same across all the table partitions
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent for snapshot {snap_dt}")
            li.append(df_partition)

        return pd.concat(li, axis = 0, ignore_index = True)
    
    def load(self, snap_dt: date, **kws) -> pyspark.sql.DataFrame:
        """Read as spark dataframe (one snapshot date) data, and can also access from sc temporary view

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_file_path_with_s3_prefix(snap_dt)
        if hasattr(self, "sc"):
            df = self.sc.read_parquet(filepath, **kws)
            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")
    
    def loads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pyspark.sql.DataFrame:
        """reads as pyspark dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        file_paths = [self.get_file_path_with_s3_prefix(snap_dt) for snap_dt in snap_dts]
        if hasattr(self, "sc"):
            df = self.sc.read_parquet(*file_paths, **kws) # spark will automatically stack all tables together
            # do the partition check
            if partition_key is not None and partition_key not in df.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found")

            # pick up random snapshot data to check the columns are consistent
            one_columns = self.load_random_snap().columns
            if len(one_columns) != len(pd.Index(one_columns).union(df.columns)):
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent in one of the snapshot data")

            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")
    
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe, could be partitions

        :param snap_dt: which snap date to delete
        :return:
        """
        filepath = self.get_default_file_path(snap_dt=snap_dt)
        if self.s3a.list_objs(filepath) is None or len(self.s3a.list_objs(filepath)) == 1:
            # it is not folder
            self.s3a.delete_obj(remote_path = filepath)
        else:
            # it is a folder, i.e., partitions
            self.s3a.delete_folder(remote_folder_path = filepath)

    def drop(self):
        """drop the whole table

        :return:
        """
        self.s3a.delete_folder(remote_folder_path = self.dir)

    def duplicate(self, dst_schema: str, dst_table: str):
        """duplicate the existing schema.table to new one, the existing one will be kept

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        new = SnapshotDataManagerS3(schema = dst_schema, table = dst_table)
        for snap_dt in tqdm(self.get_existing_snap_dts()):
            from_filename = self.get_default_file_path(snap_dt=snap_dt)
            to_filename = new.get_default_file_path(snap_dt=snap_dt)
            self.s3a.copy_obj(
                from_path = from_filename,
                to_path = to_filename
            )
        

    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date on disk

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        #key = self.s3a.get_obj(remote_path=self.get_default_file_path(snap_dt=snap_dt))
        #size_bytes = key['ContentLength'] # in kb
        size_bytes = self.s3a.list_objs(remote_path=self.get_default_file_path(snap_dt=snap_dt))['Size'].sum()
        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size
    
    def is_valid(self, snap_dt, size_threshold: float = 0.5) -> bool:
        """Check if the file is valid
        
        """
        return self.disk_space(snap_dt, unit = 'MB') > size_threshold